(*
Calibreで生成したEPUB2ファイルをEPUB3ファイルに変換するプログラム
2014.11.13 Takeshi Yoneki
ostra@asahi.email.ne.jp

確認環境
OCaml 4.02.1
OPAM 1.2.0
Ubuntu 12.04

ocamlfind ocamlopt -package batteries,camlzip,xml-light -linkpkg epub2to3_O.ml -o epub2to3_O

Ubuntuへの最新OCamlの環境構築2014.11版

基本的にはOPAMを全面採用。ただしOPAMのビルドにはOCamlが必要なので、
それはSynapticなどで入れておく。
OCamlのバージョンは多少古くても大丈夫。
build-essentialとかm4とかも必要。
他にも必要なものはあるかもしれないがエラーが出て何が足りないか教えてくれる。

OPAMのソースアーカイブを展開してビルド
(https://github.com/ocaml/opam/releases)。
(2014.11ではopam-1.2.0.tar.gz)
README.mdにどうすべきか書いてある。
./configure
make lib-ext
make
sudo make install

OPAMの初期化
opam init
.bashrcに環境変数用スクリプト呼び出しを入れて貰う。

最新OCamlインストール
opam switch
でインストール可能なバージョンが判る。
2014.11での最新は4.02.1
opam switch 4.02.1
ログインし直す。

ツールを入れる。
opam install omake

ライブラリを入れる。
opam install calmzip
opam install xml-light
opam install batteries

~/.opam/opam-init/variables.sh
CAML_LD_LIBRARY_PATH に元のシステムの（古いOCamlの）余計なパスがあり、
それがbatteriesなどの邪魔をするのでその余計なパスを削除する。

-----------------------------------------

Linux機でのepubfix運用はScala版のepub2to3だったが、F#版のepub2to3に置き換える。
無圧縮にならない問題は未確認。
Kobo向けにidの出力機能を加える。

2020.12.16

epub2to3_Oをベースにepubfix_Oを作成。
TToolsとTHtmlParserを使うように修正。
Kobo対応を移植。
Linux機でのepubfix運用をepubfix_Oに置き換える。

2021.03.03
*)

open TTools

exception ERROR_RULETRANSFORMER

type config = {
  vertical: bool ;
  src: string ;
  dst: string ;
  css: string ;
  cssText: string ;
  id: bool
}

let defaultConfig = { vertical = false; src = ""; dst = ""; css = ""; cssText = ""; id = true }

let rec parseParam args cfg =
  match args with
  | p :: v :: axs -> (
    match p with
    | "--direction" | "-r" -> parseParam axs { cfg with vertical = (String.lowercase v = "v") }
    | "--src" | "-s" | "--input" | "-i" -> parseParam axs { cfg with src = v }
    | "--dst" | "-d" | "--output" | "-o" -> parseParam axs { cfg with dst = v }
    | "--css" | "-c" -> parseParam axs { cfg with css = v }
    | "--id" | "--kobo" | "-k" -> parseParam axs { cfg with id = (String.lowercase v = "true") }
    | _ -> parseParam (v :: axs) cfg
  )
  | _ -> cfg

let hasSvgHtml inText =
  let rootE = THtmlParser.getDocumentElement (THtmlParser.parse inText) in
  THtmlParser.getElementsByTagName rootE "svg" <> []

let getSvgHtmlPathList inz =
  TList.filter_map
    (fun ent ->
      let n = String.lowercase ent.Zip.filename in
      if TString.ends_with n ".html" || TString.ends_with n ".xhtml" then
        let src = Zip.read_entry inz ent in
        if hasSvgHtml src then
          Some ent.Zip.filename
        else
          None
      else
        None
    )
    (Zip.entries inz)

let getNcxPath rootE =
  let mEs = THtmlParser.getElementsByTagName rootE "manifest" in
  if mEs <> [] then
    let h =
      TList.filter_map
        (fun e ->
          match e with
          | THtmlParser.Element (tag, _, _) ->
            if tag = "item" && THtmlParser.getAttribute e "media-type" = "application/x-dtbncx+xml" then
              Some (THtmlParser.getAttribute e "href")
            else
              None
          | _ ->
            None
        )
        (THtmlParser.getChildren (List.hd mEs))
    in
    if h <> [] then
      List.hd h
    else
      ""
  else
    ""

let getNavPathFromNcxPath ncxPath =
  let n = "nav.html" in
  if Filename.basename ncxPath = ncxPath then
    n
  else
    Filename.concat (Filename.dirname ncxPath) n

let fixOpf cfg inText svgHtmlList =
  let doc = THtmlParser.parse inText in
  let rootE = THtmlParser.getDocumentElement (THtmlParser.parse inText) in
  let ncxPath = getNcxPath rootE in
  let rec trans node path =
    match node with
    | THtmlParser.Element (tag, attr, ch) ->
      (* let n = tag in *)
      if tag = "package" then
        THtmlParser.setAttribute node "version" "3.0"
      else if TString.ends_with tag ":creator" || TString.ends_with tag "contributor" then
        THtmlParser.Element (tag, [], ch)
      else if TString.ends_with tag ":identifier" then
        let sc = TList.filter_map (fun (tag, v) -> if TString.ends_with tag ":scheme" then Some tag else None) attr in
        if List.length sc > 0 then
          THtmlParser.removeAttribute node (List.hd sc)
        else
          node
      else if TString.ends_with tag ":date" then
        let s = THtmlParser.getElementText node in
        let v = (String.sub s 0 19) ^ "Z" in
        THtmlParser.Element ("meta", [("property", "dcterms:modified")], [THtmlParser.PCData v])
      else if tag = "manifest" then
        if ncxPath <> "" then
          let h = getNavPathFromNcxPath ncxPath in
          let e = THtmlParser.Element 
            ("item", 
              [ ("href", h); 
                ("id", "nav");
                ("media-type", "application/xhtml+xml");
                ("properties", "nav")], 
             [])
          in
          THtmlParser.Element (tag, attr, e :: ch)
        else
          node
      else if tag = "spine" then
        if cfg.vertical then
          THtmlParser.setAttribute node "page-progression-direction" "rtl"
        else
          node
      else if tag = "item" then
        let h = THtmlParser.getAttribute node "href" in
        if List.mem h svgHtmlList then
          THtmlParser.setAttribute node "properties" "svg"
        else
          node
      else
        node
    | _ ->
      node
  in
  let doc2 = THtmlParser.ruleTransformer trans doc in
  let doc3 = THtmlParser.pretty "\t" doc2 in
  THtmlParser.toXmlString doc3

let convNcxToNav inText =
  let rootE = THtmlParser.getDocumentElement (THtmlParser.parse inText) in
  let title = 
    let dtEs = THtmlParser.getElementsByTagName rootE "docTitle" in
    if dtEs <> [] then
      let tEs = THtmlParser.getElementsByTagName (List.hd dtEs) "text" in
      if tEs <> [] then
        THtmlParser.getElementText (List.hd tEs)
      else
        ""
    else
      ""
  in
  let npEs = THtmlParser.getElementsByTagName rootE "navPoint" in
  let npList =
    TList.filter_map
      (fun e ->
        let npTitle = 
          let nlEs = THtmlParser.getElementsByTagName e "navLabel" in
          if nlEs <> [] then
            let tEs = THtmlParser.getElementsByTagName (List.hd nlEs) "text" in
            if tEs <> [] then
              THtmlParser.getElementText (List.hd tEs)
            else
              ""
          else
            ""
        in
        let npSrc = 
          let cEs = THtmlParser.getElementsByTagName e "content" in
          if cEs <> [] then
            THtmlParser.getAttribute (List.hd cEs) "src"
          else
            ""
        in
        if npTitle = "" || npSrc = "" then
          None
        else
          Some (npTitle, npSrc)
      )
      npEs
  in

  let tmpText = "<html xmlns=\"http://www.w3.org/1999/xhtml\" xmlns:epub=\"http://www.idpf.org/2007/ops\"><head><meta charset=\"UTF-8\" /><title /></head><body><nav epub:type=\"toc\"><h1 /><ol /></nav></body></html>" in
  let tmpDoc = THtmlParser.parse tmpText in
  (*let tmpRootE = THtmlParser.getDocumentElement (THtmlParser.parse tmpText) in*)


  let rec trans node path =
    match node with
    | THtmlParser.Element (tag, attr, children) ->
      if tag = "title" || tag = "h1" then
        THtmlParser.Element (tag, attr, [THtmlParser.PCData title])
      else if tag = "ol" then
        let cs = 
          List.map
            (fun (npTitle, npSrc) ->
              let ea = THtmlParser.Element ("a", [("href", npSrc)], [THtmlParser.PCData npTitle]) in
              let eli = THtmlParser.Element ("li", [], [ea]) in
              eli
            )
            npList
        in
        THtmlParser.Element (tag, attr, cs)
      else
        node
    | _ ->
      node
  in

  let d = THtmlParser.ruleTransformer trans tmpDoc in
  let d2 = THtmlParser.pretty "\t" d in
  THtmlParser.toXmlString d2

let fixCss cfg inText =
  let t1 = "\n\n@page { margin: 10%; }\n" in
  let t2 = "html { -epub-writing-mode: vertical-rl; writing-mode: vertical-rl; }\n" in
  let t3 = cfg.cssText in
  if cfg.vertical then
    inText ^ t1 ^ t2 ^ t3
  else
    inText ^ t1 ^ t3

let fixHtml cfg inText =
  let idNumber = ref 1 in
  let doc = THtmlParser.parse inText in
  (*let rootE = THtmlParser.getDocumentElement (THtmlParser.parse inText) in*)
  let rec trans node path =
    match node with
    | THtmlParser.Element (tag, attr, ch) -> (
      match tag with
      | "meta" ->
        if THtmlParser.getAttribute node "http-equiv" <> "" then
          THtmlParser.Empty
        else
          node
      | "link" -> 
        if THtmlParser.getAttribute node "rel" = "alternate stylesheet" then
          THtmlParser.Empty
        else
          node
      | "p" | "div" | "span" | "h1" | "h2" | "a" | "img" ->
        if cfg.id && THtmlParser.getAttribute node "id" = "" then
          let id = "kobo." ^ (string_of_int !idNumber) ^ ".1" in
          idNumber := !idNumber + 1 ;
          THtmlParser.setAttribute node "id" id
        else
          node
      | _ ->
        node
    )
    | _ ->
      node
  in
  let doc2 = THtmlParser.ruleTransformer trans doc in
  let doc3 = THtmlParser.pretty "\t" doc2 in
  THtmlParser.toXmlString doc3

let fixZip cfg inPath outPath =
  let inz = Zip.open_in inPath in
  let outz = Zip.open_out outPath in
  let svgs = getSvgHtmlPathList inz in
  let _ =
    List.iter
      (fun ent ->
        let n = String.lowercase ent.Zip.filename in
        let src = Zip.read_entry inz ent in
        if TString.ends_with n ".opf" then
          Zip.add_entry (fixOpf cfg src svgs) outz ent.Zip.filename
        else if TString.ends_with n ".ncx" then
          let _ = Zip.add_entry (convNcxToNav src) outz (getNavPathFromNcxPath ent.Zip.filename) in
          Zip.add_entry src outz ent.Zip.filename
        else if TString.ends_with n ".css" then
          Zip.add_entry (fixCss cfg src) outz ent.Zip.filename
        else if TString.ends_with n ".html" || TString.ends_with n ".xhtml" then
          Zip.add_entry (fixHtml cfg src) outz ent.Zip.filename
        else if (Filename.basename ent.Zip.filename) = "mimetype" then
          Zip.add_entry src outz ent.Zip.filename ~level:0
        else
          Zip.add_entry src outz ent.Zip.filename
      )
      (Zip.entries inz)
  in
  let _ = Zip.close_out outz in
  let _ = Zip.close_in inz in
  ()

let main () =
  let cfg = parseParam (List.tl (Array.to_list Sys.argv)) defaultConfig in
  let ct = if String.length cfg.css > 0 then TFile.read_file_all cfg.css else "" in
  let cfg = { cfg with cssText = ct } in
  let _ = fixZip cfg cfg.src cfg.dst in
  ()

let _= main ()
